# -*- coding: utf-8 -*-
import datetime

DATABASE_URL = ("https://s3.eu-central-1.amazonaws.com/"
                "chanjo/ccds.20150512.sqlite3")
BED_URL = "https://s3.eu-central-1.amazonaws.com/chanjo/ccds.20150512.bed.zip"
CCDS_DATE = datetime.date(2015, 5, 12)

ARCHIVE = {}



# -*- coding: utf-8 -*-
"""
chanjo.utils
~~~~~~~~~~~~~

A few general utility functions that might also be useful outside
Chanjo.
"""
import pkg_resources
import random
import sys

import click


def list_get(list_obj, index, default=None):
    """Like ``.get`` for list object.

    Args:
        list_obj (list): list to look up an index in
        index (int): index position to look up
        default (Optional[object]): default return value. Defaults to None.

    Returns:
        object: any object found at the index or ``default``
    """
    try:
        return list_obj[index]
    except IndexError:
        return default


def id_generator(size=8):
    """Randomly generate an id of length N (size) we can recognize.

    Think Italian or Japanese or Native American.
    Modified from: `Stackoverflow <http://stackoverflow.com/questions/2257441>`_
    and `ActiveState <http://code.activestate.com/recipes/526619/>`_.

    Args:
        size (Optional[int]): length of id, number of characters.
                              Defaults to 8.

    Returns:
        str: randomly generated string

    Examples:

    .. code-block:: python

        >>> id_generator()
        'palevedu'
        >>> id_generator(3)
        'sun'
    """
    variables = u'aeiou'
    consonants = u'bdfghklmnprstvw'

    return ''.join([random.choice(variables if i % 2 else consonants)
                    for i in range(size)])


def validate_stdin(context, param, value):
    """Validate piped input contains some data.

    Raises:
        click.BadParameter: if STDIN is empty
    """
    # check if input is a file or stdin
    if value.name == '<stdin>':
        # raise error if stdin is empty
        if sys.stdin.isatty():
            raise click.BadParameter('you need to pipe something to stdin')
    return value


class EntryPointsCLI(click.MultiCommand):

    """Add subcommands dynamically to a CLI via entry points."""

    def _iter_commands(self):
        """Iterate over all subcommands as defined by the entry point."""
        return {entry_point.name: entry_point for entry_point in
                pkg_resources.iter_entry_points('chanjo.subcommands.3')}

    def list_commands(self, ctx):
        """List the available commands."""
        commands = self._iter_commands()
        return commands.keys()

    def get_command(self, ctx, name):
        """Load one of the available commands."""
        commands = self._iter_commands()
        return commands[name].load()



# -*- coding: utf-8 -*-
import os

from chanjo.annotate.sambamba import run_sambamba


def test_run_sambamba():
    bam = 'tests/fixtures/ccds.mini.sorted.bam'
    bed = 'tests/fixtures/CCDS.mini.bed'
    out = '/tmp/ccds.coverage.bed'
    thresholds = (10, 20)
    run_sambamba(bam, bed, outfile=out, cov_treshold=thresholds)
    assert os.path.exists(out)



#!/usr/bin/env python
from chanjo.compat import text_type
from chanjo.utils import id_generator


def test_id_generator():
    """Test generating a random id."""
    # difficult to test randomly generated strings...
    assert len(id_generator()) == 8
    assert isinstance(id_generator(), text_type)

    # test with custom sized id
    assert len(id_generator(3)) == 3

    # test edge case with 0 lenght
    assert id_generator(0) == ''



# -*- coding: utf-8 -*-
import io
import json

import pytest
import yaml

from chanjo.config import Config


def test_init():
    fake_path = 'tests/fixtures/nothing.json'

    # Test lazy loading (not pointing to a config file)
    config = Config()
    assert config.markup is json

    # Test loading with some defaults
    config = Config(defaults=dict(name='P.T. Anderson'))
    assert config['name'] == 'P.T. Anderson'

    # Test loading with an alternate markup engine
    config = Config(markup=yaml)
    assert config.markup is yaml

    # Test loading with a non-existent config file
    config = Config(fake_path)
    assert config == {}

    # Test loading with an existing config file
    config = Config('tests/fixtures/something.json')
    assert config['title'] == 'Magnolia'


class TestConfig:
    def setup(self):
        self.defaults = {'type': 'movie',
                         'person': {'name': 'P.T. Anderson',
                                    'job': 'Director'}}
        self.config = Config(defaults=self.defaults)

    def test_load(self):
        # Test loading non-conflicting data
        path = 'tests/fixtures/something.json'
        with io.open(path, encoding='utf-8') as handle:
            self.config.load(handle)

        assert self.config['person']['name'] == 'P.T. Anderson'
        assert self.config['title'] == 'Magnolia'

        # Test loading conflicting data
        conflict_path = 'tests/fixtures/conflict.json'
        with io.open(conflict_path, encoding='utf-8') as handle:
            self.config.load(handle)

        assert self.config['name'] == u'Alfonso Cuarón'

        # Test loading with a non-supported markup engine
        self.config.markup = dict()
        with pytest.raises(NotImplementedError):
            with io.open(path, encoding='utf-8') as handle:
                self.config.load(handle)

        # Test loading config with bad syntax
        self.config.markup = json
        json_path = 'tests/fixtures/bad-syntax.json'
        with pytest.raises(ValueError):
            with io.open(json_path, encoding='utf-8') as handle:
                self.config.load(handle)

    def test_resolve_key(self):
        # Test simple case
        section, key = self.config._resolve_key('person.name')
        assert section == self.defaults['person']
        assert key == 'name'

        # Test with a different base
        section, key = self.config._resolve_key('person.job',
                                                base=self.config.user_data)
        assert section == {}  # 'user_data' is an empty ``dict`` in this case
        assert key == 'job'

    def test_set(self):
        # Test setting a global setting
        self.config.set('country', 'United States')
        assert self.config['country'] == 'United States'

        # Test setting a nested setting (in existing sub-group)
        self.config.set('person.name', 'Quentin Tarantino')
        assert self.config['person']['name'] == 'Quentin Tarantino'

        # Test setting a nested setting (in a new sub-group)
        self.config.set('bonus.joke', 'Knock, knock.')
        assert self.config['bonus']['joke'] == 'Knock, knock.'

        # Test multiple nested settings
        value = 'Do not ask me'
        self.config.set('what.is.this.useful.for', value)
        assert self.config['what']['is']['this']['useful']['for'] == value



# -*- coding: utf-8 -*-
"""
chanjo.questions
~~~~~~~~~~~~~~~~~~
"""
from click.termui import echo, style

# Python 2.x compatability
try:
    input = raw_input
except NameError:
    pass


def build_prompt(prompt, replacement=''):
    """Craft a prompt to ask a question.

    Provides a few modifications that should make it as simple as possible
    to ask questions.

    .. code-block:: python

        >>> build_prompt('name', '(Heisenberg)')
        'name: (Heisenberg)'
        >>> build_prompt('What is your name?', '(Heisenberg)')
        'What is your name? (Heisenberg)'
        >>> build_prompt('I {default}ed to China', '[style(walk, fg='red')]')
        'I [walk]ed to China'

    Args:
        prompt (str): base prompt, ``{default}`` will be substituted
        replacement (str, optional): string to substitute ``{default}`` with

    Returns:
        str: The full, modified prompt string
    """
    # prefix the question-prefix
    prompt = "[{}] ".format(style('?', fg='green')) + prompt

    # determine whether it seems user has tried to format the prompt
    if "{default}" not in prompt:
        # automate some formatting for the sake of convenience
        if not prompt.endswith(' '):
            if not prompt.endswith('?') or prompt.endswith(':'):
                prompt += ':'

        # add the default option to the end of prompt
        prompt += " {default}"

    # make the default-substitution and ensure an empty space at the end
    return (prompt.format(default=replacement)).rstrip() + ' '


def ask(prompt, default=None, color='cyan'):
    """Ask a question, waits for user input.

    Replacement for "input". Updates the same line by replacing "default".

    .. code-block:: python

        >>> my_name = ask('Say my name: %s ', 'Heisenberg')
        # Wait for user input
        Say my name: (Heisenberg) Walter
        # Updates *the same* line with 'Walter' in green
        Say my name: Walter
        >>> print(my_name)
        Walter

    Inspired by 'bower_ init' which confirms user input by replacing the
    default option in line (ref_).

    Args:
        prompt (str): Question to print, '%s' will be replaced by default
        default (str, optional): Default option unless replaced by user
        color (str, optional): Some common color like 'red', 'green', 'yellow'

    Returns:
        str: User input or default

    .. _bower: http://bower.io/
    .. _ref: http://stackoverflow.com/questions/12586601
    """
    # helper variables
    MOVE_CURSOR_UP = '\x1b[1A'
    ERASE_LINE = '\x1b[2K'

    # determine if a default was submitted
    if default:
        # prepare the default-part of the prompt
        default_string = "(%s)" % default
    else:
        # not relevant since ``promt`` shouldn't include a '%s'
        default_string = ''

    # pass question to user and wait for response
    # write default option in parentheses, use it as response if nothing
    # was submitted by user.
    response = input(build_prompt(prompt, default_string)) or default
    if isinstance(default, list) and isinstance(response, str):
        sep = ',' if ',' in response else None
        response = [int(item) for item in response.split(sep)]

    # print the updated confirmation line by replacing the previous
    echo(MOVE_CURSOR_UP + ERASE_LINE
         + build_prompt(prompt, style(str(response) or '', fg=color)))

    return response


def questionnaire(questions, confirm_color='cyan'):
    """Ask a set of questions á la 'bower_ init'.

    Returns a dict with responses (or defaults). Abstraction for multiple
    consecutive calls to :func:`ask`.

    Note that the function doesn't try to be smart about what types
    (e.g. 'int' or 'str') the user is typing in. You should do any
    nessesary conversions downstream.

    .. code-block:: python

        >>> questions = [('name', 'Heisenberg'), ('age', 'How old are you?', 50)]
        >>> questionnaire(questions)
        name: (Heisenberg) Walter       # -> name: Walter
        How old are you? (25)           # -> How old are you? 25
        [out] {'age': 25, 'name': 'Walter'}

    Args:
        questions (list): List of tuples on the form:
          ``(<question id>, [<prompt>], [<default>])``.
        confirm_color str, optional): Some common color like to color the
          confirm messages.

    Returns:
        dict: Each question Id is returned with either user input or the
              default value (**unmodified**)

    .. _bower: http://bower.io/
    """
    # initialize dict to store responses (or defaults)
    responses = {}

    # ask each question to sending to :func:`ask`
    for question in questions:
        # save responses with question Ids
        responses[question[0]] = ask(*question[-2:], color=confirm_color)

    return responses



# -*- coding: utf-8 -*-
from __future__ import division
import itertools
import logging

from sqlalchemy.sql import func

from chanjo.compat import itervalues
from chanjo.utils import list_get
from .core import Store
from .converter import ChanjoConverterMixin
from .models import Exon, ExonStatistic, Sample, Transcript
from .utils import filter_samples, group_by_field, predict_gender

logger = logging.getLogger(__name__)


class ChanjoAPI(Store, ChanjoConverterMixin):

    """Interface to chanjo database for asking interesting questions.

    Attributes:
        weighten_average (BinaryExpression): weighter mean for metrics
    """

    def init_app(self, app, key_base='CHANJO_'):
        """Configure API (Flask style) after lazy initialization.

        Args:
            app (Flask): Flask app instance
            key_base (str): namespace to look for under ``app.config``

        Returns:
            ChanjoAPI: ``self``
        """
        uri = app.config["{}URI".format(key_base)]
        self.connect(db_uri=uri)
        return self

    @property
    def weighted_average(self):
        """Build a dynamic field to calculate a weighted average.

        Works for any of the exon metrics.

        Returns:
            BinaryExpression: dynamic field for a query
        """
        # length of the exon
        weight = Exon.end - Exon.start
        total_weight = func.sum(weight)
        total_value = func.sum(ExonStatistic.value * weight)
        weighted_mean = (total_value / total_weight).label('weighted_mean')
        return weighted_mean

    def samples(self, group_id=None, sample_ids=None):
        """Get samples from the database.

        Args:
            group_id (Optional[str]): sample group identifier
            sample_ids (Optional[List[str]]): sample ids

        Returns:
            List[Sample]: samples objects from the database
        """
        query = self.query(Sample).order_by(Sample.sample_id)
        query = filter_samples(query, group_id=group_id, sample_ids=sample_ids)
        return query

    def means(self, query=None):
        """Calculate means on a per sample basis.

        Args:
            query (Optional[Query]): initialized and filtered query

        Returns:
            List[tuple]: sample id and grouped mean metrics
        """
        query = query or self.query()
        ready_query = (query.add_columns(Sample.sample_id,
                                         ExonStatistic.metric,
                                         self.weighted_average)
                            .join(ExonStatistic.sample, ExonStatistic.exon)
                            .group_by(Sample.sample_id, ExonStatistic.metric)
                            .order_by(Sample.sample_id))
        results = group_by_field(ready_query)
        return results

    def region_alt(self, region_id, sample_id=None, per=None):
        """Parse region id as input for `region` method.

        Args:
            region_id (str): for example 1:1242314-1243419
            sample_id (Optional[str]): filter to only a single sample
            per (Optional[str]): report on a per exon basis ("exon")

        Returns:
            dict: weighted metrics (across samples)
        """
        chromosome, pos_str = region_id.split(':')
        start, end = [int(pos) for pos in pos_str.split('-')]
        data = self.region(chromosome, start, end, sample_id=sample_id,
                           per=per)
        return data

    def region(self, chromosome, start, end, sample_id=None, per=None):
        """Report coverage across a genomics region (of exons).

        Args:
            chromosome (str): chromosome id
            start (int): start position (inclusive)
            end (int): end position (inclusive)
            sample_id (Optional[str]): filter to only a single sample
            per (Optional[str]): report on a per exon basis ("exon")

        Returns:
            dict: weighted metrics (across samples)
        """
        results = (self.query(Exon.exon_id, ExonStatistic.metric,
                              self.weighted_average)
                       .join(ExonStatistic.exon)
                       .filter(Exon.chromosome == chromosome,
                               Exon.start >= start,
                               Exon.end <= end)
                       .group_by(ExonStatistic.metric)
                       .order_by(Exon.exon_id))

        if sample_id:
            results = (results.join(ExonStatistic.sample)
                              .filter(Sample.sample_id == sample_id))

        if per == 'exon':
            results = results.group_by(Exon.exon_id)
            exon_groups = group_by_field(results)
            return exon_groups
        else:
            data = {metric: value for exon_id, metric, value in results}
            return data

    def gene(self, *gene_ids):
        r"""Report aggregate statistics for particular genes.

        Args:
            \*gene_ids (List[str]): gene ids

        Returns:
            List[dict]: metrics grouped by sample and gene
        """
        samples = {}
        for gene_id in gene_ids:
            logger.debug('figure out which transcripts the gene belongs to')
            exon_objs = self.gene_exons(gene_id).all()
            if len(exon_objs) == 0:
                raise AttributeError("gene id not in database: {}"
                                     .format(gene_id))
            exon_ids = [exon_obj.exon_id for exon_obj in exon_objs]
            query = (self.query().filter(Exon.exon_id.in_(exon_ids)))
            for sample_id, data in self.means(query):
                if sample_id not in samples:
                    samples[sample_id] = {'sample_id': sample_id, 'genes': {}}
                samples[sample_id]['genes'][gene_id] = data

        return itervalues(samples)

    def transcripts(self, *transcript_ids):
        r"""Report metrics on specific transcripts.

        Args:
            \*transcript_ids (List[str]): transcript ids

        Returns:
            List[tuple]: sample, transcript, metric, value
        """
        results = (self.query(Sample.sample_id,
                              Transcript.transcript_id,
                              ExonStatistic.metric,
                              self.weighted_average)
                       .join(ExonStatistic.sample, ExonStatistic.exon,
                             Exon.transcripts)
                       .filter(Transcript.transcript_id.in_(transcript_ids))
                       .group_by(Sample.sample_id, ExonStatistic.metric,
                                 Transcript.transcript_id))
        return results

    def incomplete_exons(self, query=None, level=10, threshold=100):
        """Fetch exons with incomplete coverage at a specifc level.

        Args:
            level (Optional[int]): completeness level, default: 10
            threshold (Optional[int]): % completeness cutoff, default: 100
            group_id (Optional[str]): sample group identifier
            sample_ids (Optional[List[str]]): sample ids

        Returns:
            List[tuple]: sample_id, dict with exons and completeness
        """
        completeness = "completeness_{}".format(level)
        query = query or self.query()
        ready_query = (self.query(Sample.sample_id,
                                  Exon.exon_id,
                                  ExonStatistic.value)
                           .join(ExonStatistic.exon, ExonStatistic.sample)
                           .filter(ExonStatistic.metric == completeness,
                                   ExonStatistic.value < threshold)
                           .order_by(Sample.sample_id))
        exon_samples = group_by_field(ready_query)
        return exon_samples

    def sex_coverage(self, sex_chromosomes=('X', 'Y')):
        """Query for average on X/Y chromsomes.

        Args:
            sex_chromosomes (Optional[tuple]): chromosome ids

        Returns:
            List[tuple]: sample, chromosome, weighted average coverage
        """
        query = (self.query(Sample.sample_id,
                            Exon.chromosome,
                            self.weighted_average)
                     .join(ExonStatistic.exon, ExonStatistic.sample)
                     .filter(Exon.chromosome.in_(sex_chromosomes),
                             ExonStatistic.metric == 'mean_coverage')
                     .group_by(Sample.sample_id, Exon.chromosome))
        return query

    def sex_check(self, group_id=None, sample_ids=None):
        """Predict gender based on coverage of sex chromosomes.

        Args:
            group_id (Optional[str]): sample group identifier
            sample_ids (Optional[List[str]]): sample ids

        Returns:
            tuple: sample, gender, coverage X, coverage, Y
        """
        query = self.sex_coverage()
        query = filter_samples(query, group_id=group_id, sample_ids=sample_ids)
        logger.debug('group rows based on sample')
        samples = itertools.groupby(query, lambda row: row[0])
        for sample_id, chromosomes in samples:
            sex_coverage = [coverage for _, _, coverage in chromosomes]
            logger.debug('predict gender')
            # run the predictor
            gender = predict_gender(*sex_coverage)
            yield sample_id, gender, sex_coverage[0], sex_coverage[1]

    def diagnostic_yield(self, sample_id, exon_ids=None, query=None, level=10,
                         threshold=100):
        """Calculate transcripts that aren't completely covered.

        This metric only applies to one sample in isolation. Otherwise
        it's hard to know what to do with exons that are covered or
        not covered across multiple samples.

        Args:
            sample_id (str): unique sample id
        """
        if exon_ids:
            query = self.query().filter(Exon.exon_id.in_(exon_ids))
        else:
            query = self.query()

        level = "completeness_{}".format(level)

        all_query = (query.add_columns(Transcript.transcript_id)
                          .distinct(Transcript.transcript_id)
                          .join(Exon.transcripts))
        all_count = all_query.count()

        logger.debug('find out which exons failed')
        yield_query = (query.add_columns(Exon.exon_id)
                            .join(ExonStatistic.sample, ExonStatistic.exon)
                            .filter(ExonStatistic.metric == level,
                                    ExonStatistic.value < threshold,
                                    Sample.sample_id == sample_id))
        exons = [row[0] for row in yield_query]
        transcripts = self.exon_transcripts(exons)

        tx_count = transcripts.count()
        diagnostic_yield = 100 - (tx_count/all_count * 100)
        return {
            "diagnostic_yield": diagnostic_yield,
            "count": tx_count,
            "total_count": all_count,
            "transcripts": transcripts
        }

    def completeness_levels(self):
        """Return a list of included completeness levels."""
        metrics = (self.query(ExonStatistic.metric)
                       .distinct(ExonStatistic.metric))
        # for all completeness levels, extract the level as int and full name
        levels = ((int(field[0].split('_')[-1]), field[0])
                  for field in metrics
                  if field[0].startswith('completeness'))
        # sort them based on the level int
        sorted_levels = sorted(levels, key=lambda level: list_get(level, 0))
        return sorted_levels



# -*- coding: utf-8 -*-
import logging
import os

import click
from path import path

from chanjo import bootstrap as bootstrap_api
from .init import init

logger = logging.getLogger(__name__)


@click.command()
@click.option('-f', '--force', is_flag=True)
@click.argument('directory', default='./coverage', required=False)
@click.pass_context
def bootstrap(context, force, directory):
    """Link related genomic elements."""
    bootstrap_api.pull(directory, force=force)

    config_path = os.path.join(directory, 'chanjo.yaml')
    context.obj.config_path = path(config_path)
    context.invoke(init, automate=True)



# -*- coding: utf-8 -*-
import logging

import click
from click.termui import echo, style

import chanjo
from chanjo.compat import iteritems
from chanjo.config import questionnaire
from chanjo.store import Store

logger = logging.getLogger(__name__)


@click.command()
@click.option('-s', '--setup', is_flag=True, help='setup database tables')
@click.option('-r', '--reset', is_flag=True, help='reset an existing database')
@click.option('-a', '--automate', is_flag=True, help='run non-interactively')
@click.pass_context
def init(context, setup, reset, automate):
    """Walk user through setting up a new config file."""
    # print a nice welcome message
    click.echo(chanjo.__banner__)

    cov_tresholds = (context.obj.get('sambamba', {})
                                .get('cov_treshold', [10, 20]))
    defaults = {'sambamba.cov_treshold': {'value': cov_tresholds,
                                          'prompt': 'sufficient coverage'},
                'database': {'value': context.obj['database'],
                             'prompt': 'central database path/URI'}}

    if not automate:
        questions = [(key, value['prompt'], value['value'])
                     for key, value in iteritems(defaults)]
        # launch init pipeline
        init_pipeline('chanjo', context.obj, questions)
    else:
        logger.info('setting default config values')
        for key, value in iteritems(defaults):
            context.obj.set(key, value['value'], scope=context.obj.user_data)

    # write to the config file
    context.obj.save(default_flow_style=False)

    if setup:
        chanjo_db = Store(uri=context.obj.user_data['database'])
        if reset:
            chanjo_db.tear_down()
        chanjo_db.set_up()


def init_pipeline(program, config, questions):
    """Initializes a config object by interactively asking questions."""
    if config.user_data:
        # Some existing user settings were found, warn about overwriting them
        message = "{program} {note}\tThe existing {file} will be updated"
        segments = dict(program=program, note=style('existing', fg='yellow'),
                        file=style(config.config_path.basename(), fg='white'))
        echo(message.format(**segments))

    # Launch questionnaire
    user_defaults = questionnaire(questions)
    # Set the selected user defaults
    for dot_key, value in user_defaults.items():
        config.set(dot_key, value, scope=config.user_data)



# -*- coding: utf-8 -*-
import pytest

from chanjo.exc import BedFormattingError
from chanjo.parse import bed

TEST_OUTPUT = 'tests/fixtures/test.sambamba.bed'


def test_extra_fields():
    columns = ['X-CCDS43917.1,X-CCDS55362.1,X-CCDS55363.1',
               'X-ASMTL,X-ASMTL,X-ASMTL']
    combos = [combo for combo in bed.extra_fields(columns)]
    assert len(combos) == 3
    assert combos[0] == ('X-CCDS43917.1', 'X-ASMTL')


def test_chanjo():
    with open(TEST_OUTPUT, 'r') as handle:
        data_rows = [data for data in bed.chanjo(handle)]

    data = data_rows[0]
    assert data['name'] == '1-69090-70007'
    assert data['score'] == 0
    assert data['strand'] == '+'
    assert list(data['elements']) == [('CCDS30547.1', 'OR4F5')]


def test_chanj_with_sambamba():
    bed_lines = ["# chrom\tchromStart\tchromEnd\treadCount\tmeanCoverage"
                    "\tsampleName",
                 "1\t69089\t70007\t232\t25.4946\tADM992A10\t"]

    with pytest.raises(BedFormattingError):
        list(bed.chanjo(bed_lines))



# -*- coding: utf-8 -*-
from chanjo.config import remove_ansi
from chanjo.config.questions import build_prompt


def test_build_prompt():
    # Test simple prompt
    prompt = '[?] name: (Heisenberg) '
    assert remove_ansi(build_prompt('name', '(Heisenberg)')) == prompt

    # Test custom prompt
    prompt = '[?] What is your name? (Heisenberg) '
    built_prompt = build_prompt('What is your name?', '(Heisenberg)')
    assert prompt == remove_ansi(built_prompt)

    # Test highly customized prompt (default not in the end)
    prompt = '[?] I [walk]ed to China '
    built_prompt = build_prompt('I {default}ed to China', '[walk]')
    assert prompt == remove_ansi(built_prompt)



# -*- coding: utf-8 -*-
from chanjo.store.utils import predict_gender


def test_predict_gender():
    assert predict_gender(x_coverage=10.2, y_coverage=8.1) == 'male'
    assert predict_gender(x_coverage=20.9, y_coverage=.01) == 'female'
    # shouldn't raise ``ZeroDivisionError``
    assert predict_gender(x_coverage=5.12, y_coverage=0) == 'female'

    # what does this even mean?
    assert predict_gender(x_coverage=0, y_coverage=10) == 'unknown'



# -*- coding: utf-8 -*-
"""
chanjo.config
~~~~~~~~~~~~~~

Config module with Click integration.
"""
import yaml as markup

from .core import Config
from .questions import ask, build_prompt, questionnaire
from .utils import remove_ansi

CONFIG_FILE_NAME = ("{program}.{extension}"
                    .format(program='chanjo', extension=markup.__name__))



# -*- coding: utf-8 -*-
import logging

import click
from sqlalchemy.exc import IntegrityError

from chanjo.load.sambamba import rows as sambamba_rows
from chanjo.parse import sambamba
from chanjo.store import Store
from chanjo.utils import validate_stdin

logger = logging.getLogger(__name__)


@click.command()
@click.option('-s', '--sample', help='override sample id from file')
@click.option('-g', '--group', help='id to group related samples')
@click.argument('bed_stream', callback=validate_stdin,
                type=click.File(encoding='utf-8'), default='-', required=False)
@click.pass_context
def load(context, sample, group, bed_stream):
    """Load Sambamba output into the database for a sample."""
    chanjo_db = Store(uri=context.obj['database'])
    try:
        load_sambamba(chanjo_db, bed_stream, sample_id=sample, group_id=group)
    except IntegrityError:
        logger.error('sample already loaded, rolling back')
        chanjo_db.session.rollback()
        context.abort()


def load_sambamba(chanjo_db, bed_iterable, sample_id=None, group_id=None):
    """Load Sambamba BED output from a stream."""
    rows = sambamba.depth_output(bed_iterable)
    stats = sambamba_rows(chanjo_db.session, rows, sample_id=sample_id,
                          group_id=group_id)
    for index, stat in enumerate(stats):
        chanjo_db.add(stat)
        if index % 10000 == 0:
            chanjo_db.save()
            logger.debug("processed %s annotations", index)

    chanjo_db.save()
    logger.debug("processed %s annotations", index)



# -*- coding: utf-8 -*-
"""
chanjo.cli
~~~~~~~~~~~

Command line interface (console entry points). Based on Click_.

.. _Click: http://click.pocoo.org/
"""
import logging
import click

from chanjo.compat import text_type
from chanjo.config import Config, CONFIG_FILE_NAME, markup
from chanjo.log import init_log, LEVELS
from chanjo.utils import EntryPointsCLI

from chanjo import __version__

logger = logging.getLogger(__name__)


def print_version(ctx, param, value):
    """Callback function for printing version and exiting
    Args:
        ctx (object) : Current context
        param (object) : Click parameter(s)
        value (boolean) : Click parameter was supplied or not
    Returns:
        None
    """
    if not value or ctx.resilient_parsing:
        return
    click.echo("chanjo version: {0}".format(__version__))
    ctx.exit()

@click.group(cls=EntryPointsCLI)
@click.option('-c', '--config',
                default=CONFIG_FILE_NAME,
                type=click.Path(),
                help='path to config file'
)
@click.option('-d', '--database',
                type=text_type,
                help='path/URI of the SQL database'
)
@click.option('-v', '--verbose',
                count=True,
                default=1,
                help="Increase output verbosity. Can be used multiple times, eg. -vv"
)
@click.option('--log_file',
                type=click.Path()
)
@click.option('--version',
                is_flag=True,
                callback=print_version,
                expose_value=False,
                is_eager=True
)
@click.pass_context
def root(context, config, database, verbose, log_file):
    """Clinical sequencing coverage analysis tool."""
    # setup logging
    loglevel = LEVELS.get(min(verbose, 2), 'WARNING')
    init_log(logging.getLogger(), loglevel=loglevel, filename=log_file)
    logger.info("version {0}".format(__version__))

    # avoid setting global defaults in Click options, do it below when
    # updating the config object
    context.obj = Config(config, markup=markup)
    context.obj['database'] = (database or
                               context.obj.get('database', 'coverage.sqlite3'))

    # update the context with new defaults from the config file
    context.default_map = context.obj
